# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for VertexAI models."""

import os
import unittest
from unittest import mock

from google.auth import exceptions
import langfun.core as lf
from langfun.core.llms import vertexai


class VertexAITest(unittest.TestCase):
  """Tests for Vertex model with REST API."""

  @mock.patch.object(vertexai.VertexAI, 'credentials', new=True)
  def test_project_and_location_check(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `project`'):
      _ = vertexai.VertexAIGeminiPro1()._api_initialized

    with self.assertRaisesRegex(ValueError, 'Please specify `location`'):
      _ = vertexai.VertexAIGeminiPro1(project='abc')._api_initialized

    self.assertTrue(
        vertexai.VertexAIGeminiPro1(
            project='abc', location='us-central1'
        )._api_initialized
    )

    os.environ['VERTEXAI_PROJECT'] = 'abc'
    os.environ['VERTEXAI_LOCATION'] = 'us-central1'
    model = vertexai.VertexAIGeminiPro1()
    self.assertTrue(model.model_id.startswith('VertexAI('))
    self.assertIn('us-central1', model.api_endpoint)
    self.assertTrue(model._api_initialized)
    self.assertIsNotNone(model.session())
    del os.environ['VERTEXAI_PROJECT']
    del os.environ['VERTEXAI_LOCATION']


class VertexAIAnthropicTest(unittest.TestCase):
  """Tests for VertexAI Anthropic models."""

  def test_basics(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `project`'):
      lm = vertexai.VertexAIClaude3_5_Sonnet_20241022()
      lm('hi')

    model = vertexai.VertexAIClaude3_5_Sonnet_20241022(project='langfun')

    # NOTE(daiyip): For OSS users, default credentials are not available unless
    # users have already set up their GCP project. Therefore we ignore the
    # exception here.
    try:
      model._initialize()
    except exceptions.DefaultCredentialsError:
      pass

    self.assertEqual(
        model.api_endpoint,
        (
            'https://us-east5-aiplatform.googleapis.com/v1/projects/'
            'langfun/locations/us-east5/publishers/anthropic/'
            'models/claude-3-5-sonnet-v2@20241022:streamRawPredict'
        )
    )
    self.assertEqual(
        model.headers,
        {
            'Content-Type': 'application/json; charset=utf-8',
        },
    )
    request = model.request(
        lf.UserMessage('hi'), lf.LMSamplingOptions(temperature=0.0),
    )
    self.assertEqual(
        request,
        {
            'anthropic_version': 'vertex-2023-10-16',
            'max_tokens': 8192,
            'messages': [
                {'content': [{'text': 'hi', 'type': 'text'}], 'role': 'user'}
            ],
            'stream': False,
            'temperature': 0.0,
            'top_k': 40,
        },
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Vertex AI generative models."""

import os
from typing import Annotated, Any, Literal

import langfun.core as lf
from langfun.core.llms import anthropic
from langfun.core.llms import gemini
from langfun.core.llms import openai_compatible
from langfun.core.llms import rest
import pyglove as pg

try:
  # pylint: disable=g-import-not-at-top
  from google import auth as google_auth
  from google.auth import credentials as credentials_lib
  from google.auth.transport import requests as auth_requests
  # pylint: enable=g-import-not-at-top

  Credentials = credentials_lib.Credentials
except ImportError:
  google_auth = None
  credentials_lib = None
  auth_requests = None
  Credentials = Any


@pg.use_init_args(['api_endpoint'])
class VertexAI(rest.REST):
  """Base class for VertexAI models.

  This class handles the authentication of vertex AI models. Subclasses
  should implement `request` and `result` methods, as well as the `api_endpoint`
  property. Or let users to provide them as __init__ arguments.

  Please check out VertexAIGemini in `gemini.py` as an example.
  """

  model: Annotated[
      str | None,
      'Model ID.'
  ] = None

  project: Annotated[
      str | None,
      (
          'Vertex AI project ID. Or set from environment variable '
          'VERTEXAI_PROJECT.'
      ),
  ] = None

  location: Annotated[
      str | None,
      (
          'Vertex AI service location. Or set from environment variable '
          'VERTEXAI_LOCATION.'
      ),
  ] = None

  credentials: Annotated[
      Credentials | None,
      (
          'Credentials to use. If None, the default credentials to the '
          'environment will be used.'
      ),
  ] = None

  def _on_bound(self):
    super()._on_bound()
    if google_auth is None:
      raise ValueError(
          'Please install "langfun[llm-google-vertex]" to use Vertex AI models.'
      )
    self._project = None
    self._credentials = None

  def _initialize(self):
    project = self.project or os.environ.get('VERTEXAI_PROJECT', None)
    if not project:
      raise ValueError(
          'Please specify `project` during `__init__` or set environment '
          'variable `VERTEXAI_PROJECT` with your Vertex AI project ID.'
      )

    location = self.location or os.environ.get('VERTEXAI_LOCATION', None)
    if not location:
      raise ValueError(
          'Please specify `location` during `__init__` or set environment '
          'variable `VERTEXAI_LOCATION` with your Vertex AI service location.'
      )

    self._project = project
    self._location = location

    credentials = self.credentials
    if credentials is None:
      # Use default credentials.
      credentials, _ = google_auth.default(
          scopes=['https://www.googleapis.com/auth/cloud-platform']
      )
    self._credentials = credentials

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return f'VertexAI({self.model})'

  def _session(self):
    assert self._credentials is not None
    assert auth_requests is not None
    return auth_requests.AuthorizedSession(self._credentials)


#
# Gemini models served by Vertex AI.
#


@pg.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class VertexAIGemini(VertexAI, gemini.Gemini):
  """Gemini models served by Vertex AI.."""

  @property
  def api_endpoint(self) -> str:
    assert self._api_initialized
    return (
        f'https://{self._location}-aiplatform.googleapis.com/v1/projects/'
        f'{self._project}/locations/{self._location}/publishers/google/'
        f'models/{self.model}:generateContent'
    )


class VertexAIGemini2Flash(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 model launched on 02/05/2025."""

  api_version = 'v1beta'
  model = 'gemini-2.0-flash-001'
  location = 'us-central1'


class VertexAIGemini2ProExp_20250205(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 Pro model launched on 02/05/2025."""

  api_version = 'v1beta'
  model = 'gemini-2.0-pro-exp-02-05'
  location = 'us-central1'


class VertexAIGemini2FlashThinkingExp_20250121(VertexAIGemini):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 Thinking model launched on 01/21/2025."""

  api_version = 'v1beta'
  model = 'gemini-2.0-flash-thinking-exp-01-21'
  timeout = None
  location = 'us-central1'


class VertexAIGeminiFlash2_0ThinkingExp_20241219(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini Flash 2.0 Thinking model launched on 12/19/2024."""

  api_version = 'v1alpha'
  model = 'gemini-2.0-flash-thinking-exp-1219'
  timeout = None


class VertexAIGeminiFlash2_0Exp(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 2.0 Flash model."""

  model = 'gemini-2.0-flash-exp'


class VertexAIGeminiExp_20241206(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini Experimental model launched on 12/06/2024."""

  model = 'gemini-exp-1206'


class VertexAIGeminiExp_20241114(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini Experimental model launched on 11/14/2024."""

  model = 'gemini-exp-1114'


class VertexAIGeminiPro1_5(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model."""

  model = 'gemini-1.5-pro-002'
  location = 'us-central1'


class VertexAIGeminiPro1_5_002(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model."""

  model = 'gemini-1.5-pro-002'
  location = 'us-central1'


class VertexAIGeminiPro1_5_001(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Pro model."""

  model = 'gemini-1.5-pro-001'
  location = 'us-central1'


class VertexAIGeminiFlash1_5(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model."""

  model = 'gemini-1.5-flash'
  location = 'us-central1'


class VertexAIGeminiFlash1_5_002(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model."""

  model = 'gemini-1.5-flash-002'
  location = 'us-central1'


class VertexAIGeminiFlash1_5_001(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.5 Flash model."""

  model = 'gemini-1.5-flash-001'
  location = 'us-central1'


class VertexAIGeminiPro1(VertexAIGemini):  # pylint: disable=invalid-name
  """Vertex AI Gemini 1.0 Pro model."""

  model = 'gemini-1.0-pro'


#
# Anthropic models on Vertex AI.
#


@pg.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class VertexAIAnthropic(VertexAI, anthropic.Anthropic):
  """Anthropic models on VertexAI."""

  location: Annotated[
      Literal['us-east5', 'europe-west1'],
      'GCP location with Anthropic models hosted.'
  ] = 'us-east5'

  api_version = 'vertex-2023-10-16'

  @property
  def headers(self):
    return {
        'Content-Type': 'application/json; charset=utf-8',
    }

  @property
  def api_endpoint(self) -> str:
    return (
        f'https://{self.location}-aiplatform.googleapis.com/v1/projects/'
        f'{self._project}/locations/{self.location}/publishers/anthropic/'
        f'models/{self.model}:streamRawPredict'
    )

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ):
    request = super().request(prompt, sampling_options)
    request['anthropic_version'] = self.api_version
    del request['model']
    return request


# pylint: disable=invalid-name


class VertexAIClaude3_Opus_20240229(VertexAIAnthropic):
  """Anthropic's Claude 3 Opus model on VertexAI."""
  model = 'claude-3-opus@20240229'


class VertexAIClaude3_5_Sonnet_20241022(VertexAIAnthropic):
  """Anthropic's Claude 3.5 Sonnet model on VertexAI."""
  model = 'claude-3-5-sonnet-v2@20241022'


class VertexAIClaude3_5_Sonnet_20240620(VertexAIAnthropic):
  """Anthropic's Claude 3.5 Sonnet model on VertexAI."""
  model = 'claude-3-5-sonnet@20240620'


class VertexAIClaude3_5_Haiku_20241022(VertexAIAnthropic):
  """Anthropic's Claude 3.5 Haiku model on VertexAI."""
  model = 'claude-3-5-haiku@20241022'

# pylint: enable=invalid-name

#
# Llama models on Vertex AI.
# pylint: disable=line-too-long
# Pricing: https://cloud.google.com/vertex-ai/generative-ai/pricing?_gl=1*ukuk6u*_ga*MjEzMjc4NjM2My4xNzMzODg4OTg3*_ga_WH2QY8WWF5*MTczNzEzNDU1Mi4xMjQuMS4xNzM3MTM0NzczLjU5LjAuMA..#meta-models
# pylint: enable=line-too-long

LLAMA_MODELS = {
    'llama-3.2-90b-vision-instruct-maas': pg.Dict(
        latest_update='2024-09-25',
        in_service=True,
        rpm=0,
        tpm=0,
        # Free during preview.
        cost_per_1m_input_tokens=None,
        cost_per_1m_output_tokens=None,
    ),
    'llama-3.1-405b-instruct-maas': pg.Dict(
        latest_update='2024-09-25',
        in_service=True,
        rpm=0,
        tpm=0,
        # GA.
        cost_per_1m_input_tokens=5,
        cost_per_1m_output_tokens=16,
    ),
    'llama-3.1-70b-instruct-maas': pg.Dict(
        latest_update='2024-09-25',
        in_service=True,
        rpm=0,
        tpm=0,
        # Free during preview.
        cost_per_1m_input_tokens=None,
        cost_per_1m_output_tokens=None,
    ),
    'llama-3.1-8b-instruct-maas': pg.Dict(
        latest_update='2024-09-25',
        in_service=True,
        rpm=0,
        tpm=0,
        # Free during preview.
        cost_per_1m_input_tokens=None,
        cost_per_1m_output_tokens=None,
    )
}


@pg.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class VertexAILlama(VertexAI, openai_compatible.OpenAICompatible):
  """Llama models on VertexAI."""

  model: pg.typing.Annotated[
      pg.typing.Enum(pg.MISSING_VALUE, list(LLAMA_MODELS.keys())),
      'Llama model ID.',
  ]

  locations: Annotated[
      Literal['us-central1'],
      (
          'GCP locations with Llama models hosted. '
          'See https://cloud.google.com/vertex-ai/generative-ai/docs/partner-models/llama#regions-quotas'
      )
  ] = 'us-central1'

  @property
  def api_endpoint(self) -> str:
    assert self._api_initialized
    return (
        f'https://{self._location}-aiplatform.googleapis.com/v1beta1/projects/'
        f'{self._project}/locations/{self._location}/endpoints/'
        f'openapi/chat/completions'
    )

  def request(
      self,
      prompt: lf.Message,
      sampling_options: lf.LMSamplingOptions
  ):
    request = super().request(prompt, sampling_options)
    request['model'] = f'meta/{self.model}'
    return request

  @property
  def max_concurrency(self) -> int:
    rpm = LLAMA_MODELS[self.model].get('rpm', 0)
    tpm = LLAMA_MODELS[self.model].get('tpm', 0)
    return self.rate_to_max_concurrency(
        requests_per_min=rpm, tokens_per_min=tpm
    )

  def estimate_cost(
      self,
      num_input_tokens: int,
      num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    cost_per_1m_input_tokens = LLAMA_MODELS[self.model].get(
        'cost_per_1m_input_tokens', None
    )
    cost_per_1m_output_tokens = LLAMA_MODELS[self.model].get(
        'cost_per_1m_output_tokens', None
    )
    if cost_per_1m_output_tokens is None or cost_per_1m_input_tokens is None:
      return None
    return (
        cost_per_1m_input_tokens * num_input_tokens
        + cost_per_1m_output_tokens * num_output_tokens
    ) / 1000_000


# pylint: disable=invalid-name
class VertexAILlama3_2_90B(VertexAILlama):
  """Llama 3.2 90B vision instruct model on VertexAI."""

  model = 'llama-3.2-90b-vision-instruct-maas'


class VertexAILlama3_1_405B(VertexAILlama):
  """Llama 3.1 405B vision instruct model on VertexAI."""

  model = 'llama-3.1-405b-instruct-maas'


class VertexAILlama3_1_70B(VertexAILlama):
  """Llama 3.1 70B vision instruct model on VertexAI."""

  model = 'llama-3.1-70b-instruct-maas'


class VertexAILlama3_1_8B(VertexAILlama):
  """Llama 3.1 8B vision instruct model on VertexAI."""

  model = 'llama-3.1-8b-instruct-maas'
# pylint: enable=invalid-name

#
# Mistral models on Vertex AI.
# pylint: disable=line-too-long
# Pricing: https://cloud.google.com/vertex-ai/generative-ai/pricing?_gl=1*ukuk6u*_ga*MjEzMjc4NjM2My4xNzMzODg4OTg3*_ga_WH2QY8WWF5*MTczNzEzNDU1Mi4xMjQuMS4xNzM3MTM0NzczLjU5LjAuMA..#mistral-models
# pylint: enable=line-too-long


MISTRAL_MODELS = {
    'mistral-large-2411': pg.Dict(
        latest_update='2024-11-21',
        in_service=True,
        rpm=0,
        tpm=0,
        # GA.
        cost_per_1m_input_tokens=2,
        cost_per_1m_output_tokens=6,
    ),
    'mistral-large@2407': pg.Dict(
        latest_update='2024-07-24',
        in_service=True,
        rpm=0,
        tpm=0,
        # GA.
        cost_per_1m_input_tokens=2,
        cost_per_1m_output_tokens=6,
    ),
    'mistral-nemo@2407': pg.Dict(
        latest_update='2024-07-24',
        in_service=True,
        rpm=0,
        tpm=0,
        # GA.
        cost_per_1m_input_tokens=0.15,
        cost_per_1m_output_tokens=0.15,
    ),
    'codestral-2501': pg.Dict(
        latest_update='2025-01-13',
        in_service=True,
        rpm=0,
        tpm=0,
        # GA.
        cost_per_1m_input_tokens=0.3,
        cost_per_1m_output_tokens=0.9,
    ),
    'codestral@2405': pg.Dict(
        latest_update='2024-05-29',
        in_service=True,
        rpm=0,
        tpm=0,
        # GA.
        cost_per_1m_input_tokens=0.2,
        cost_per_1m_output_tokens=0.6,
    ),
}


@pg.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class VertexAIMistral(VertexAI, openai_compatible.OpenAICompatible):
  """Mistral AI models on VertexAI."""

  model: pg.typing.Annotated[
      pg.typing.Enum(pg.MISSING_VALUE, list(MISTRAL_MODELS.keys())),
      'Mistral model ID.',
  ]

  locations: Annotated[
      Literal['us-central1', 'europe-west4'],
      (
          'GCP locations with Mistral models hosted. '
          'See https://cloud.google.com/vertex-ai/generative-ai/docs/partner-models/mistral#regions-quotas'
      )
  ] = 'us-central1'

  @property
  def api_endpoint(self) -> str:
    assert self._api_initialized
    return (
        f'https://{self._location}-aiplatform.googleapis.com/v1/projects/'
        f'{self._project}/locations/{self._location}/publishers/mistralai/'
        f'models/{self.model}:rawPredict'
    )

  @property
  def max_concurrency(self) -> int:
    rpm = MISTRAL_MODELS[self.model].get('rpm', 0)
    tpm = MISTRAL_MODELS[self.model].get('tpm', 0)
    return self.rate_to_max_concurrency(
        requests_per_min=rpm, tokens_per_min=tpm
    )

  def estimate_cost(
      self,
      num_input_tokens: int,
      num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    cost_per_1m_input_tokens = MISTRAL_MODELS[self.model].get(
        'cost_per_1m_input_tokens', None
    )
    cost_per_1m_output_tokens = MISTRAL_MODELS[self.model].get(
        'cost_per_1m_output_tokens', None
    )
    if cost_per_1m_output_tokens is None or cost_per_1m_input_tokens is None:
      return None
    return (
        cost_per_1m_input_tokens * num_input_tokens
        + cost_per_1m_output_tokens * num_output_tokens
    ) / 1000_000


# pylint: disable=invalid-name
class VertexAIMistralLarge_20241121(VertexAIMistral):
  """Mistral Large model on VertexAI released on 2024/11/21."""

  model = 'mistral-large-2411'


class VertexAIMistralLarge_20240724(VertexAIMistral):
  """Mistral Large model on VertexAI released on 2024/07/24."""

  model = 'mistral-large@2407'


class VertexAIMistralNemo_20240724(VertexAIMistral):
  """Mistral Nemo model on VertexAI released on 2024/07/24."""

  model = 'mistral-nemo@2407'


class VertexAICodestral_20250113(VertexAIMistral):
  """Mistral Nemo model on VertexAI released on 2024/07/24."""

  model = 'codestral-2501'


class VertexAICodestral_20240529(VertexAIMistral):
  """Mistral Nemo model on VertexAI released on 2024/05/29."""

  model = 'codestral@2405'
# pylint: enable=invalid-name



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gemini REST API (Shared by Google GenAI and Vertex AI)."""

import base64
from typing import Any

import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.llms import rest
import pyglove as pg

# Supported modalities.

IMAGE_TYPES = [
    'image/png',
    'image/jpeg',
    'image/webp',
    'image/heic',
    'image/heif',
]

AUDIO_TYPES = [
    'audio/aac',
    'audio/flac',
    'audio/mp3',
    'audio/m4a',
    'audio/mpeg',
    'audio/mpga',
    'audio/mp4',
    'audio/opus',
    'audio/pcm',
    'audio/wav',
    'audio/webm',
]

VIDEO_TYPES = [
    'video/mov',
    'video/mpeg',
    'video/mpegps',
    'video/mpg',
    'video/mp4',
    'video/webm',
    'video/wmv',
    'video/x-flv',
    'video/3gpp',
    'video/quicktime',
]

DOCUMENT_TYPES = [
    'application/pdf',
    'text/plain',
    'text/csv',
    'text/html',
    'text/xml',
    'text/x-script.python',
    'application/json',
]

TEXT_ONLY = []

ALL_MODALITIES = (
    IMAGE_TYPES + AUDIO_TYPES + VIDEO_TYPES + DOCUMENT_TYPES
)

SUPPORTED_MODELS_AND_SETTINGS = {
    # For automatically rate control and cost estimation, we explicitly register
    # supported models here. This may be inconvenient for new models, but it
    # helps us to keep track of the models and their pricing.
    # Models and RPM are from
    # https://ai.google.dev/gemini-api/docs/models/gemini?_gl=1*114hbho*_up*MQ..&gclid=Cj0KCQiAst67BhCEARIsAKKdWOljBY5aQdNQ41zOPkXFCwymUfMNFl_7ukm1veAf75ZTD9qWFrFr11IaApL3EALw_wcB
    # Pricing in US dollars, from https://ai.google.dev/pricing
    # as of 2025-01-03.
    # NOTE: Please update google_genai.py, vertexai.py, __init__.py when
    # adding new models.
    # !!! PLEASE KEEP MODELS SORTED BY RELEASE DATE !!!
    'gemini-2.0-flash-001': pg.Dict(
        latest_update='2025-02-05',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10_000,
        tpm_free=40_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0,
        cost_per_1m_output_tokens_up_to_128k=0,
        cost_per_1m_cached_tokens_up_to_128k=0,
        cost_per_1m_input_tokens_longer_than_128k=0,
        cost_per_1m_output_tokens_longer_than_128k=0,
        cost_per_1m_cached_tokens_longer_than_128k=0,
    ),
    'gemini-2.0-flash': pg.Dict(
        latest_update='2025-02-05',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10_000,
        tpm_free=40_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0.1,
        cost_per_1m_output_tokens_up_to_128k=0.4,
        cost_per_1m_cached_tokens_up_to_128k=0.025,
        cost_per_1m_input_tokens_longer_than_128k=0.1,
        cost_per_1m_output_tokens_longer_than_128k=0.4,
        cost_per_1m_cached_tokens_longer_than_128k=0.025,
    ),
    'gemini-2.0-pro-exp-02-05': pg.Dict(
        latest_update='2025-02-05',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10,
        tpm_free=4_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0.1,
        cost_per_1m_output_tokens_up_to_128k=0.4,
        cost_per_1m_cached_tokens_up_to_128k=0.025,
        cost_per_1m_input_tokens_longer_than_128k=0.1,
        cost_per_1m_output_tokens_longer_than_128k=0.4,
        cost_per_1m_cached_tokens_longer_than_128k=0.025,
    ),
    'gemini-2.0-flash-thinking-exp-01-21': pg.Dict(
        latest_update='2024-01-21',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10,
        tpm_free=40_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0,
        cost_per_1m_output_tokens_up_to_128k=0,
        cost_per_1m_cached_tokens_up_to_128k=0,
        cost_per_1m_input_tokens_longer_than_128k=0,
        cost_per_1m_output_tokens_longer_than_128k=0,
        cost_per_1m_cached_tokens_longer_than_128k=0,
    ),
    'gemini-2.0-flash-thinking-exp-1219': pg.Dict(
        latest_update='2024-12-19',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10,
        tpm_free=4_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0,
        cost_per_1m_output_tokens_up_to_128k=0,
        cost_per_1m_cached_tokens_up_to_128k=0,
        cost_per_1m_input_tokens_longer_than_128k=0,
        cost_per_1m_output_tokens_longer_than_128k=0,
        cost_per_1m_cached_tokens_longer_than_128k=0,
    ),
    'gemini-2.0-flash-exp': pg.Dict(
        latest_update='2024-12-11',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10,
        tpm_free=4_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0,
        cost_per_1m_output_tokens_up_to_128k=0,
        cost_per_1m_cached_tokens_up_to_128k=0,
        cost_per_1m_input_tokens_longer_than_128k=0,
        cost_per_1m_output_tokens_longer_than_128k=0,
        cost_per_1m_cached_tokens_longer_than_128k=0,
    ),
    'gemini-exp-1206': pg.Dict(
        latest_update='2024-12-06',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10,
        tpm_free=4_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0,
        cost_per_1m_output_tokens_up_to_128k=0,
        cost_per_1m_cached_tokens_up_to_128k=0,
        cost_per_1m_input_tokens_longer_than_128k=0,
        cost_per_1m_output_tokens_longer_than_128k=0,
        cost_per_1m_cached_tokens_longer_than_128k=0,
    ),
    'learnlm-1.5-pro-experimental': pg.Dict(
        latest_update='2024-11-19',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10,
        tpm_free=4_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0,
        cost_per_1m_output_tokens_up_to_128k=0,
        cost_per_1m_cached_tokens_up_to_128k=0,
        cost_per_1m_input_tokens_longer_than_128k=0,
        cost_per_1m_output_tokens_longer_than_128k=0,
        cost_per_1m_cached_tokens_longer_than_128k=0,
    ),
    'gemini-exp-1114': pg.Dict(
        latest_update='2024-11-14',
        experimental=True,
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=10,
        tpm_free=4_000_000,
        rpm_paid=0,
        tpm_paid=0,
        cost_per_1m_input_tokens_up_to_128k=0,
        cost_per_1m_output_tokens_up_to_128k=0,
        cost_per_1m_cached_tokens_up_to_128k=0,
        cost_per_1m_input_tokens_longer_than_128k=0,
        cost_per_1m_output_tokens_longer_than_128k=0,
        cost_per_1m_cached_tokens_longer_than_128k=0,
    ),
    'gemini-1.5-flash-latest': pg.Dict(
        latest_update='2024-09-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=15,
        tpm_free=1_000_000,
        rpm_paid=2000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=0.075,
        cost_per_1m_output_tokens_up_to_128k=0.3,
        cost_per_1m_cached_tokens_up_to_128k=0.01875,
        cost_per_1m_input_tokens_longer_than_128k=0.15,
        cost_per_1m_output_tokens_longer_than_128k=0.6,
        cost_per_1m_cached_tokens_longer_than_128k=0.0375,
    ),
    'gemini-1.5-flash': pg.Dict(
        latest_update='2024-09-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=15,
        tpm_free=1_000_000,
        rpm_paid=2000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=0.075,
        cost_per_1m_output_tokens_up_to_128k=0.3,
        cost_per_1m_cached_tokens_up_to_128k=0.01875,
        cost_per_1m_input_tokens_longer_than_128k=0.15,
        cost_per_1m_output_tokens_longer_than_128k=0.6,
        cost_per_1m_cached_tokens_longer_than_128k=0.0375,
    ),
    'gemini-1.5-flash-001': pg.Dict(
        latest_update='2024-09-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=15,
        tpm_free=1_000_000,
        rpm_paid=2000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=0.075,
        cost_per_1m_output_tokens_up_to_128k=0.3,
        cost_per_1m_cached_tokens_up_to_128k=0.01875,
        cost_per_1m_input_tokens_longer_than_128k=0.15,
        cost_per_1m_output_tokens_longer_than_128k=0.6,
        cost_per_1m_cached_tokens_longer_than_128k=0.0375,
    ),
    'gemini-1.5-flash-002': pg.Dict(
        latest_update='2024-09-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=15,
        tpm_free=1_000_000,
        rpm_paid=2000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=0.075,
        cost_per_1m_output_tokens_up_to_128k=0.3,
        cost_per_1m_cached_tokens_up_to_128k=0.01875,
        cost_per_1m_input_tokens_longer_than_128k=0.15,
        cost_per_1m_output_tokens_longer_than_128k=0.6,
        cost_per_1m_cached_tokens_longer_than_128k=0.0375,
    ),
    'gemini-1.5-flash-8b': pg.Dict(
        latest_update='2024-10-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=15,
        tpm_free=1_000_000,
        rpm_paid=4000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=0.0375,
        cost_per_1m_output_tokens_up_to_128k=0.15,
        cost_per_1m_cached_tokens_up_to_128k=0.01,
        cost_per_1m_input_tokens_longer_than_128k=0.075,
        cost_per_1m_output_tokens_longer_than_128k=0.3,
        cost_per_1m_cached_tokens_longer_than_128k=0.02,
    ),
    'gemini-1.5-flash-8b-001': pg.Dict(
        latest_update='2024-10-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=15,
        tpm_free=1_000_000,
        rpm_paid=4000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=0.0375,
        cost_per_1m_output_tokens_up_to_128k=0.15,
        cost_per_1m_cached_tokens_up_to_128k=0.01,
        cost_per_1m_input_tokens_longer_than_128k=0.075,
        cost_per_1m_output_tokens_longer_than_128k=0.3,
        cost_per_1m_cached_tokens_longer_than_128k=0.02,
    ),
    'gemini-1.5-pro-latest': pg.Dict(
        latest_update='2024-09-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=2,
        tpm_free=32_000,
        rpm_paid=1000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=1.25,
        cost_per_1m_output_tokens_up_to_128k=5.00,
        cost_per_1m_cached_tokens_up_to_128k=0.3125,
        cost_per_1m_input_tokens_longer_than_128k=2.5,
        cost_per_1m_output_tokens_longer_than_128k=10.00,
        cost_per_1m_cached_tokens_longer_than_128k=0.625,
    ),
    'gemini-1.5-pro': pg.Dict(
        latest_update='2024-09-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=2,
        tpm_free=32_000,
        rpm_paid=1000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=1.25,
        cost_per_1m_output_tokens_up_to_128k=5.00,
        cost_per_1m_cached_tokens_up_to_128k=0.3125,
        cost_per_1m_input_tokens_longer_than_128k=2.5,
        cost_per_1m_output_tokens_longer_than_128k=10.00,
        cost_per_1m_cached_tokens_longer_than_128k=0.625,
    ),
    'gemini-1.5-pro-001': pg.Dict(
        latest_update='2024-09-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=2,
        tpm_free=32_000,
        rpm_paid=1000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=1.25,
        cost_per_1m_output_tokens_up_to_128k=5.00,
        cost_per_1m_cached_tokens_up_to_128k=0.3125,
        cost_per_1m_input_tokens_longer_than_128k=2.5,
        cost_per_1m_output_tokens_longer_than_128k=10.00,
        cost_per_1m_cached_tokens_longer_than_128k=0.625,
    ),
    'gemini-1.5-pro-002': pg.Dict(
        latest_update='2024-09-30',
        in_service=True,
        supported_modalities=ALL_MODALITIES,
        rpm_free=2,
        tpm_free=32_000,
        rpm_paid=1000,
        tpm_paid=4_000_000,
        cost_per_1m_input_tokens_up_to_128k=1.25,
        cost_per_1m_output_tokens_up_to_128k=5.00,
        cost_per_1m_cached_tokens_up_to_128k=0.3125,
        cost_per_1m_input_tokens_longer_than_128k=2.5,
        cost_per_1m_output_tokens_longer_than_128k=10.00,
        cost_per_1m_cached_tokens_longer_than_128k=0.625,
    ),
    'gemini-1.0-pro': pg.Dict(
        in_service=False,
        supported_modalities=TEXT_ONLY,
        rpm_free=15,
        tpm_free=32_000,
        rpm_paid=360,
        tpm_paid=120_000,
        cost_per_1m_input_tokens_up_to_128k=0.5,
        cost_per_1m_output_tokens_up_to_128k=1.5,
        cost_per_1m_cached_tokens_up_to_128k=0,
        cost_per_1m_input_tokens_longer_than_128k=0.5,
        cost_per_1m_output_tokens_longer_than_128k=1.5,
        cost_per_1m_cached_tokens_longer_than_128k=0,
    ),
}


@pg.use_init_args(['model'])
class Gemini(rest.REST):
  """Language models provided by Google GenAI."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, list(SUPPORTED_MODELS_AND_SETTINGS.keys())
      ),
      'The name of the model to use.',
  ]

  @property
  def supported_modalities(self) -> list[str]:
    """Returns the list of supported modalities."""
    return SUPPORTED_MODELS_AND_SETTINGS[self.model].supported_modalities

  @property
  def max_concurrency(self) -> int:
    """Returns the maximum number of concurrent requests."""
    return self.rate_to_max_concurrency(
        requests_per_min=max(
            SUPPORTED_MODELS_AND_SETTINGS[self.model].rpm_free,
            SUPPORTED_MODELS_AND_SETTINGS[self.model].rpm_paid
        ),
        tokens_per_min=max(
            SUPPORTED_MODELS_AND_SETTINGS[self.model].tpm_free,
            SUPPORTED_MODELS_AND_SETTINGS[self.model].tpm_paid,
        ),
    )

  def estimate_cost(
      self,
      num_input_tokens: int,
      num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    entry = SUPPORTED_MODELS_AND_SETTINGS[self.model]
    if num_input_tokens < 128_000:
      cost_per_1m_input_tokens = entry.cost_per_1m_input_tokens_up_to_128k
      cost_per_1m_output_tokens = entry.cost_per_1m_output_tokens_up_to_128k
    else:
      cost_per_1m_input_tokens = entry.cost_per_1m_input_tokens_longer_than_128k
      cost_per_1m_output_tokens = (
          entry.cost_per_1m_output_tokens_longer_than_128k
      )
    return (
        cost_per_1m_input_tokens * num_input_tokens
        + cost_per_1m_output_tokens * num_output_tokens
    ) / 1000_000

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return self.model

  @classmethod
  def dir(cls):
    return [k for k, v in SUPPORTED_MODELS_AND_SETTINGS.items() if v.in_service]

  @property
  def headers(self):
    return {
        'Content-Type': 'application/json; charset=utf-8',
    }

  def request(
      self, prompt: lf.Message, sampling_options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    request = dict(
        generationConfig=self._generation_config(prompt, sampling_options)
    )
    request['contents'] = [self._content_from_message(prompt)]
    return request

  def _generation_config(
      self, prompt: lf.Message, options: lf.LMSamplingOptions
  ) -> dict[str, Any]:
    """Returns a dict as generation config for prompt and LMSamplingOptions."""
    config = dict(
        temperature=options.temperature,
        maxOutputTokens=options.max_tokens,
        candidateCount=options.n,
        topK=options.top_k,
        topP=options.top_p,
        stopSequences=options.stop,
        seed=options.random_seed,
        responseLogprobs=options.logprobs,
        logprobs=options.top_logprobs,
    )

    if json_schema := prompt.metadata.get('json_schema'):
      if not isinstance(json_schema, dict):
        raise ValueError(
            f'`json_schema` must be a dict, got {json_schema!r}.'
        )
      json_schema = pg.to_json(json_schema)
      config['responseSchema'] = json_schema
      config['responseMimeType'] = 'application/json'
      prompt.metadata.formatted_text = (
          prompt.text
          + '\n\n [RESPONSE FORMAT (not part of prompt)]\n'
          + pg.to_json_str(json_schema, json_indent=2)
      )
    return config

  def _content_from_message(self, prompt: lf.Message) -> dict[str, Any]:
    """Gets generation content from langfun message."""
    parts = []
    for lf_chunk in prompt.chunk():
      if isinstance(lf_chunk, str):
        parts.append({'text': lf_chunk})
      elif isinstance(lf_chunk, lf_modalities.Mime):
        try:
          modalities = lf_chunk.make_compatible(
              self.supported_modalities + ['text/plain']
          )
          if isinstance(modalities, lf_modalities.Mime):
            modalities = [modalities]
          for modality in modalities:
            if modality.is_text:
              parts.append({'text': modality.to_text()})
            else:
              parts.append({
                  'inlineData': {
                      'data': base64.b64encode(modality.to_bytes()).decode(),
                      'mimeType': modality.mime_type,
                  }
              })
        except lf.ModalityError as e:
          raise lf.ModalityError(f'Unsupported modality: {lf_chunk!r}') from e
      else:
        raise lf.ModalityError(f'Unsupported modality: {lf_chunk!r}')
    return dict(role='user', parts=parts)

  def result(self, json: dict[str, Any]) -> lf.LMSamplingResult:
    messages = [
        self._message_from_content_parts(candidate['content']['parts'])
        for candidate in json['candidates']
    ]
    usage = json['usageMetadata']
    input_tokens = usage['promptTokenCount']
    output_tokens = usage['candidatesTokenCount']
    return lf.LMSamplingResult(
        [lf.LMSample(message) for message in messages],
        usage=lf.LMSamplingUsage(
            prompt_tokens=input_tokens,
            completion_tokens=output_tokens,
            total_tokens=input_tokens + output_tokens,
            estimated_cost=self.estimate_cost(
                num_input_tokens=input_tokens,
                num_output_tokens=output_tokens,
            ),
        ),
    )

  def _message_from_content_parts(
      self, parts: list[dict[str, Any]]
  ) -> lf.Message:
    """Converts Vertex AI's content parts protocol to message."""
    chunks = []
    thought_chunks = []
    for part in parts:
      if text_part := part.get('text'):
        if part.get('thought'):
          thought_chunks.append(text_part)
        else:
          chunks.append(text_part)
      else:
        raise ValueError(f'Unsupported part: {part}')
    message = lf.AIMessage.from_chunks(chunks)
    if thought_chunks:
      message.set('thought', lf.AIMessage.from_chunks(thought_chunks))
    return message



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from OpenAI."""

import os
from typing import Annotated, Any

import langfun.core as lf
from langfun.core.llms import openai_compatible
import pyglove as pg


# From https://platform.openai.com/settings/organization/limits
_DEFAULT_TPM = 250000
_DEFAULT_RPM = 3000

SUPPORTED_MODELS_AND_SETTINGS = {
    # Models from https://platform.openai.com/docs/models
    # RPM is from https://platform.openai.com/docs/guides/rate-limits
    # o1 (preview) models.
    # Pricing in US dollars, from https://openai.com/api/pricing/
    # as of 2024-10-10.
    'o3-mini-2025-01-31': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.0011,
        cost_per_1k_output_tokens=0.0044,
    ),
    'o3-mini': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.0011,
        cost_per_1k_output_tokens=0.0044,
    ),
    'o1': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.015,
        cost_per_1k_output_tokens=0.06,
    ),
    'o1-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.015,
        cost_per_1k_output_tokens=0.06,
    ),
    'o1-preview-2024-09-12': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.015,
        cost_per_1k_output_tokens=0.06,
    ),
    'o1-mini': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.012,
    ),
    'o1-mini-2024-09-12': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.012,
    ),
    # GPT-4o models
    'gpt-4o-mini': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.00015,
        cost_per_1k_output_tokens=0.0006,
    ),
    'gpt-4o-mini-2024-07-18': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.00015,
        cost_per_1k_output_tokens=0.0006,
    ),
    'gpt-4o': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.0025,
        cost_per_1k_output_tokens=0.01,
    ),
    'gpt-4o-2024-11-20': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.0025,
        cost_per_1k_output_tokens=0.01,
    ),
    'gpt-4o-2024-08-06': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.0025,
        cost_per_1k_output_tokens=0.01,
    ),
    'gpt-4o-2024-05-13': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=5000000,
        cost_per_1k_input_tokens=0.005,
        cost_per_1k_output_tokens=0.015,
    ),
    # GPT-4-Turbo models
    'gpt-4-turbo': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-turbo-2024-04-09': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-turbo-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-0125-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-1106-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-vision-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    'gpt-4-1106-vision-preview': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.01,
        cost_per_1k_output_tokens=0.03,
    ),
    # GPT-4 models
    'gpt-4': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.03,
        cost_per_1k_output_tokens=0.06,
    ),
    'gpt-4-0613': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.03,
        cost_per_1k_output_tokens=0.06,
    ),
    'gpt-4-0314': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.03,
        cost_per_1k_output_tokens=0.06,
    ),
    'gpt-4-32k': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.06,
        cost_per_1k_output_tokens=0.12,
    ),
    'gpt-4-32k-0613': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.06,
        cost_per_1k_output_tokens=0.12,
    ),
    'gpt-4-32k-0314': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=300000,
        cost_per_1k_input_tokens=0.06,
        cost_per_1k_output_tokens=0.12,
    ),
    # GPT-3.5-Turbo models
    'gpt-3.5-turbo': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.0005,
        cost_per_1k_output_tokens=0.0015,
    ),
    'gpt-3.5-turbo-0125': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.0005,
        cost_per_1k_output_tokens=0.0015,
    ),
    'gpt-3.5-turbo-1106': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.001,
        cost_per_1k_output_tokens=0.002,
    ),
    'gpt-3.5-turbo-0613': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.0015,
        cost_per_1k_output_tokens=0.002,
    ),
    'gpt-3.5-turbo-0301': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.0015,
        cost_per_1k_output_tokens=0.002,
    ),
    'gpt-3.5-turbo-16k': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.004,
    ),
    'gpt-3.5-turbo-16k-0613': pg.Dict(
        in_service=True,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.004,
    ),
    'gpt-3.5-turbo-16k-0301': pg.Dict(
        in_service=False,
        rpm=10000,
        tpm=2000000,
        cost_per_1k_input_tokens=0.003,
        cost_per_1k_output_tokens=0.004,
    ),
    # GPT-3.5 models
    'text-davinci-003': pg.Dict(
        in_service=False, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM
    ),
    'text-davinci-002': pg.Dict(
        in_service=False, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM
    ),
    'code-davinci-002': pg.Dict(
        in_service=False, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM
    ),
    # GPT-3 instruction-tuned models (Deprecated)
    'text-curie-001': pg.Dict(
        in_service=False, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM
    ),
    'text-babbage-001': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM,
    ),
    'text-ada-001': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM,
    ),
    'davinci': pg.Dict(
        in_service=False,
        rpm=_DEFAULT_RPM,
        tpm=_DEFAULT_TPM,
    ),
    'curie': pg.Dict(in_service=False, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM),
    'babbage': pg.Dict(in_service=False, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM),
    'ada': pg.Dict(in_service=False, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM),
    # GPT-3 base models that are still in service.
    'babbage-002': pg.Dict(in_service=True, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM),
    'davinci-002': pg.Dict(in_service=True, rpm=_DEFAULT_RPM, tpm=_DEFAULT_TPM),
}


@lf.use_init_args(['model'])
class OpenAI(openai_compatible.OpenAICompatible):
  """OpenAI model."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, list(SUPPORTED_MODELS_AND_SETTINGS.keys())
      ),
      'The name of the model to use.',
  ]

  api_endpoint: str = 'https://api.openai.com/v1/chat/completions'

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'OPENAI_API_KEY'."
      ),
  ] = None

  organization: Annotated[
      str | None,
      (
          'Organization. If None, the key will be read from environment '
          "variable 'OPENAI_ORGANIZATION'. Based on the value, usages from "
          "these API requests will count against the organization's quota. "
      ),
  ] = None

  project: Annotated[
      str | None,
      (
          'Project. If None, the key will be read from environment '
          "variable 'OPENAI_PROJECT'. Based on the value, usages from "
          "these API requests will count against the project's quota. "
      ),
  ] = None

  def _on_bound(self):
    super()._on_bound()
    self._api_key = None
    self._organization = None
    self._project = None

  def _initialize(self):
    api_key = self.api_key or os.environ.get('OPENAI_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `OPENAI_API_KEY` with your OpenAI API key.'
      )
    self._api_key = api_key
    self._organization = self.organization or os.environ.get(
        'OPENAI_ORGANIZATION', None
    )
    self._project = self.project or os.environ.get('OPENAI_PROJECT', None)

  @property
  def headers(self) -> dict[str, Any]:
    assert self._api_initialized
    headers = super().headers
    headers['Authorization'] = f'Bearer {self._api_key}'
    if self._organization:
      headers['OpenAI-Organization'] = self._organization
    if self._project:
      headers['OpenAI-Project'] = self._project
    return headers

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return f'OpenAI({self.model})'

  @property
  def max_concurrency(self) -> int:
    rpm = SUPPORTED_MODELS_AND_SETTINGS[self.model].get('rpm', 0)
    tpm = SUPPORTED_MODELS_AND_SETTINGS[self.model].get('tpm', 0)
    return self.rate_to_max_concurrency(
        requests_per_min=rpm, tokens_per_min=tpm
    )

  def estimate_cost(
      self,
      num_input_tokens: int,
      num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    cost_per_1k_input_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_input_tokens', None
    )
    cost_per_1k_output_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_output_tokens', None
    )
    if cost_per_1k_output_tokens is None or cost_per_1k_input_tokens is None:
      return None
    return (
        cost_per_1k_input_tokens * num_input_tokens
        + cost_per_1k_output_tokens * num_output_tokens
    ) / 1000

  @classmethod
  def dir(cls):
    return [k for k, v in SUPPORTED_MODELS_AND_SETTINGS.items() if v.in_service]

  def _request_args(
      self, options: lf.LMSamplingOptions) -> dict[str, Any]:
    # Reasoning models (o1 series) does not support `logprobs` by 2024/09/12.
    if options.logprobs and self.model.startswith(('o1-', 'o3-')):
      raise RuntimeError('`logprobs` is not supported on {self.model!r}.')
    return super()._request_args(options)


class GptO3Mini(OpenAI):
  """GPT-O3-mini."""

  model = 'o3-mini'
  multimodal = True
  timeout = None


class GptO1(OpenAI):
  """GPT-O1."""

  model = 'o1'
  multimodal = True
  timeout = None


class GptO1Preview(OpenAI):
  """GPT-O1."""
  model = 'o1-preview'
  timeout = None


class GptO1Preview_20240912(OpenAI):   # pylint: disable=invalid-name
  """GPT O1."""
  model = 'o1-preview-2024-09-12'
  timeout = None


class GptO1Mini(OpenAI):
  """GPT O1-mini."""
  model = 'o1-mini'
  timeout = None


class GptO1Mini_20240912(OpenAI):   # pylint: disable=invalid-name
  """GPT O1-mini."""
  model = 'o1-mini-2024-09-12'
  timeout = None


class Gpt4(OpenAI):
  """GPT-4."""
  model = 'gpt-4'


class Gpt4Turbo(Gpt4):
  """GPT-4 Turbo with 128K context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo'
  multimodal = True


class Gpt4Turbo_20240409(Gpt4Turbo):  # pylint:disable=invalid-name
  """GPT-4 Turbo with 128K context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo-2024-04-09'
  multimodal = True


class Gpt4TurboPreview(Gpt4):
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-turbo-preview'


class Gpt4TurboPreview_20240125(Gpt4TurboPreview):  # pylint: disable=invalid-name
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Dec. 2023."""
  model = 'gpt-4-0125-preview'


class Gpt4TurboPreview_20231106(Gpt4TurboPreview):  # pylint: disable=invalid-name
  """GPT-4 Turbo Preview with 128k context window. Knowledge up to Apr. 2023."""
  model = 'gpt-4-1106-preview'


class Gpt4VisionPreview(Gpt4):
  """GPT-4 Turbo vision preview. 128k context window. Knowledge to Apr. 2023."""
  model = 'gpt-4-vision-preview'
  multimodal = True


class Gpt4VisionPreview_20231106(Gpt4):  # pylint: disable=invalid-name
  """GPT-4 Turbo vision preview. 128k context window. Knowledge to Apr. 2023."""
  model = 'gpt-4-1106-vision-preview'


class Gpt4_20230613(Gpt4):    # pylint:disable=invalid-name
  """GPT-4 @20230613. 8K context window. Knowledge up to 9-2021."""
  model = 'gpt-4-0613'


class Gpt4_32K(Gpt4):       # pylint:disable=invalid-name
  """Latest GPT-4 with 32K context window."""
  model = 'gpt-4-32k'


class Gpt4_32K_20230613(Gpt4_32K):    # pylint:disable=invalid-name
  """GPT-4 @20230613. 32K context window. Knowledge up to 9-2021."""
  model = 'gpt-4-32k-0613'


class Gpt4oMini(OpenAI):
  """GPT-4o Mini."""
  model = 'gpt-4o-mini'
  multimodal = True


class Gpt4oMini_20240718(OpenAI):  # pylint:disable=invalid-name
  """GPT-4o Mini."""
  model = 'gpt-4o-mini-2024-07-18'
  multimodal = True


class Gpt4o(OpenAI):
  """GPT-4o."""
  model = 'gpt-4o'
  multimodal = True


class Gpt4o_20241120(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-11-20."""
  model = 'gpt-4o-2024-11-20'
  multimodal = True


class Gpt4o_20240806(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-08-06."""
  model = 'gpt-4o-2024-08-06'
  multimodal = True


class Gpt4o_20240513(OpenAI):     # pylint:disable=invalid-name
  """GPT-4o version 2024-05-13."""
  model = 'gpt-4o-2024-05-13'
  multimodal = True


class Gpt35(OpenAI):
  """GPT-3.5. 4K max tokens, trained up on data up to Sep, 2021."""
  model = 'text-davinci-003'


class Gpt35Turbo(Gpt35):
  """Most capable GPT-3.5 model, 10x cheaper than GPT35 (text-davinci-003)."""
  model = 'gpt-3.5-turbo'


class Gpt35Turbo_20240125(Gpt35Turbo):   # pylint:disable=invalid-name
  """GPT-3.5 Turbo @20240125. 16K context window. Knowledge up to 09/2021."""
  model = 'gpt-3.5-turbo-0125'


class Gpt35Turbo_20231106(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gpt3.5 Turbo @20231106. 16K context window. Knowledge up to 09/2021."""
  model = 'gpt-3.5-turbo-1106'


class Gpt35Turbo_20230613(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gpt3.5 Turbo snapshot at 2023/06/13, with 4K context window size."""
  model = 'gpt-3.5-turbo-0613'


class Gpt35Turbo16K(Gpt35Turbo):
  """Latest GPT-3.5 model with 16K context window size."""
  model = 'gpt-3.5-turbo-16k'


class Gpt35Turbo16K_20230613(Gpt35Turbo):   # pylint:disable=invalid-name
  """Gtp 3.5 Turbo 16K 0613."""
  model = 'gpt-3.5-turbo-16k-0613'


class Gpt3(OpenAI):
  """Most capable GPT-3 model (Davinci) 2K context window size.

  All GPT3 models have 2K max tokens and trained on data up to Oct 2019.
  """
  model = 'davinci'


class Gpt3Curie(Gpt3):
  """Very capable, but faster and lower cost than Davici."""
  model = 'curie'


class Gpt3Babbage(Gpt3):
  """Capable of straightforward tasks, very fast and low cost."""
  model = 'babbage'


class Gpt3Ada(Gpt3):
  """Capable of very simple tasks, the fastest/lowest cost among GPT3 models."""
  model = 'ada'



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from Google GenAI."""

import os
from typing import Annotated, Literal

import langfun.core as lf
from langfun.core.llms import gemini
import pyglove as pg


@lf.use_init_args(['model'])
@pg.members([('api_endpoint', pg.typing.Str().freeze(''))])
class GenAI(gemini.Gemini):
  """Language models provided by Google GenAI."""

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'GOOGLE_API_KEY'. "
          'Get an API key at https://ai.google.dev/tutorials/setup'
      ),
  ] = None

  api_version: Annotated[
      Literal['v1beta', 'v1alpha'],
      'The API version to use.'
  ] = 'v1beta'

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return f'GenAI({self.model})'

  @property
  def api_endpoint(self) -> str:
    api_key = self.api_key or os.environ.get('GOOGLE_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `GOOGLE_API_KEY` with your Google Cloud API key. '
          'Check out '
          'https://cloud.google.com/api-keys/docs/create-manage-api-keys '
          'for more details.'
      )
    return (
        f'https://generativelanguage.googleapis.com/{self.api_version}'
        f'/models/{self.model}:generateContent?'
        f'key={api_key}'
    )


class Gemini2Flash(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 model launched on 02/05/2025."""

  api_version = 'v1beta'
  model = 'gemini-2.0-flash'


class Gemini2ProExp_20250205(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 Pro model launched on 02/05/2025."""

  api_version = 'v1beta'
  model = 'gemini-2.0-pro-exp-02-05'


class Gemini2FlashThinkingExp_20250121(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 Thinking model launched on 01/21/2025."""

  api_version = 'v1beta'
  model = 'gemini-2.0-flash-thinking-exp-01-21'
  timeout = None


class GeminiFlash2_0ThinkingExp_20241219(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 Thinking model launched on 12/19/2024."""

  api_version = 'v1beta'
  model = 'gemini-2.0-flash-thinking-exp-1219'
  timeout = None


class GeminiFlash2_0Exp(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash 2.0 model launched on 12/11/2024."""

  model = 'gemini-2.0-flash-exp'


class GeminiExp_20241206(GenAI):  # pylint: disable=invalid-name
  """Gemini Experimental model launched on 12/06/2024."""

  model = 'gemini-exp-1206'


class GeminiExp_20241114(GenAI):  # pylint: disable=invalid-name
  """Gemini Experimental model launched on 11/14/2024."""

  model = 'gemini-exp-1114'


class GeminiPro1_5(GenAI):  # pylint: disable=invalid-name
  """Gemini Pro latest model."""

  model = 'gemini-1.5-pro-latest'


class GeminiPro1_5_002(GenAI):  # pylint: disable=invalid-name
  """Gemini Pro latest model."""

  model = 'gemini-1.5-pro-002'


class GeminiPro1_5_001(GenAI):  # pylint: disable=invalid-name
  """Gemini Pro latest model."""

  model = 'gemini-1.5-pro-001'


class GeminiFlash1_5(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash latest model."""

  model = 'gemini-1.5-flash-latest'


class GeminiFlash1_5_002(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash 1.5 model stable version 002."""

  model = 'gemini-1.5-flash-002'


class GeminiFlash1_5_001(GenAI):  # pylint: disable=invalid-name
  """Gemini Flash 1.5 model stable version 001."""

  model = 'gemini-1.5-flash-001'


class GeminiPro1(GenAI):  # pylint: disable=invalid-name
  """Gemini 1.0 Pro model."""

  model = 'gemini-1.0-pro'



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Language models from DeepSeek."""

import os
from typing import Annotated, Any

import langfun.core as lf
from langfun.core.llms import openai_compatible
import pyglove as pg

SUPPORTED_MODELS_AND_SETTINGS = {
    # pylint: disable=g-line-too-long
    # TODO(yifenglu): The RPM and TPM are arbitrary numbers. Update them once DeepSeek provides concrete guidelines.
    # DeepSeek doesn't control the rate limit at the moment: https://api-docs.deepseek.com/quick_start/rate_limit
    # The cost is based on: https://api-docs.deepseek.com/quick_start/pricing
    'deepseek-reasoner': pg.Dict(
        in_service=True,
        rpm=100,
        tpm=1000000,
        cost_per_1k_input_tokens=0.00055,
        cost_per_1k_output_tokens=0.00219,
    ),
    'deepseek-chat': pg.Dict(
        in_service=True,
        rpm=100,
        tpm=1000000,
        cost_per_1k_input_tokens=0.00014,
        cost_per_1k_output_tokens=0.00028,
    ),
}


# DeepSeek API uses an API format compatible with OpenAI.
# Reference: https://api-docs.deepseek.com/
@lf.use_init_args(['model'])
class DeepSeek(openai_compatible.OpenAICompatible):
  """DeepSeek model."""

  model: pg.typing.Annotated[
      pg.typing.Enum(
          pg.MISSING_VALUE, list(SUPPORTED_MODELS_AND_SETTINGS.keys())
      ),
      'The name of the model to use.',
  ]

  api_endpoint: str = 'https://api.deepseek.com/chat/completions'

  api_key: Annotated[
      str | None,
      (
          'API key. If None, the key will be read from environment variable '
          "'DEEPSEEK_API_KEY'."
      ),
  ] = None

  @property
  def headers(self) -> dict[str, Any]:
    api_key = self.api_key or os.environ.get('DEEPSEEK_API_KEY', None)
    if not api_key:
      raise ValueError(
          'Please specify `api_key` during `__init__` or set environment '
          'variable `DEEPSEEK_API_KEY` with your DeepSeek API key.'
      )
    headers = super().headers
    headers.update({
        'Authorization': f'Bearer {api_key}',
    })
    return headers

  @property
  def model_id(self) -> str:
    """Returns a string to identify the model."""
    return f'DeepSeek({self.model})'

  @property
  def max_concurrency(self) -> int:
    rpm = SUPPORTED_MODELS_AND_SETTINGS[self.model].get('rpm', 0)
    tpm = SUPPORTED_MODELS_AND_SETTINGS[self.model].get('tpm', 0)
    return self.rate_to_max_concurrency(
        requests_per_min=rpm, tokens_per_min=tpm
    )

  def estimate_cost(
      self, num_input_tokens: int, num_output_tokens: int
  ) -> float | None:
    """Estimate the cost based on usage."""
    cost_per_1k_input_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_input_tokens', None
    )
    cost_per_1k_output_tokens = SUPPORTED_MODELS_AND_SETTINGS[self.model].get(
        'cost_per_1k_output_tokens', None
    )
    if cost_per_1k_output_tokens is None or cost_per_1k_input_tokens is None:
      return None
    return (
        cost_per_1k_input_tokens * num_input_tokens
        + cost_per_1k_output_tokens * num_output_tokens
    ) / 1000

  @classmethod
  def dir(cls):
    return [k for k, v in SUPPORTED_MODELS_AND_SETTINGS.items() if v.in_service]


class DeepSeekReasoner(DeepSeek):
  """DeepSeek Reasoner model.

  Currently it is powered by DeepSeek-R1 model, 64k input context, 8k max
  output, 32k max CoT output.
  """

  model = 'deepseek-reasoner'


class DeepSeekChat(DeepSeek):
  """DeepSeek Chat model.

  Currently, it is powered by DeepSeek-V3 model, 64K input contenxt window and
  8k max output tokens.
  """

  model = 'deepseek-chat'
